@@ -5,8 +5,8 @@ functionality can be used via this Python interface.
 """
 
 from ._basixcpp import __version__
-from . import cell, finite_element, lattice, quadrature, variants
+from . import cell, finite_element, lattice, polynomials, quadrature, sobolev_spaces, variants
 from ._basixcpp import (CellType, LatticeType, LatticeSimplexMethod, ElementFamily, LagrangeVariant,
-                        DPCVariant, QuadratureType, PolynomialType, MapType)
+                        DPCVariant, QuadratureType, PolynomialType, MapType, SobolevSpace)
 from ._basixcpp import (create_lattice, create_element, compute_interpolation_operator, topology,
                         geometry, make_quadrature, index, tabulate_polynomials, create_custom_element)
